#include "Chiplet_Interface.h"
#include <iostream>

// Constructor implementation
ChipletInterface::ChipletInterface(sc_core::sc_module_name name, int num_pes): sc_module(name), num_pes(num_pes) {
    sc_report_handler::set_actions(C_LOG, SC_INFO, SC_LOG|SC_DISPLAY); //saves the reports to the file chiplet_comm.log
    for (int i = 0; i < num_pes; ++i) {
        // Register b_transport method for each PE receive socket
        oec_receive_sockets[i].register_b_transport(this, &ChipletInterface::oec_b_transport);   
    }
    // Register b_transport method for sockets connecting to Interposer interface
    interposer_SingleChiplet_receive_socket.register_b_transport(this, &ChipletInterface::ii_SingleChiplet_receive_b_transport);
    interposer_CrossChiplet_receive_socket.register_b_transport(this, &ChipletInterface::ii_CrossChiplet_receive_b_transport);
}

// Handle incoming transactions from the Interposer Interface
void ChipletInterface::ii_SingleChiplet_receive_b_transport(tlm_generic_payload& trans, sc_time& delay) {
    PESendPayload* orig_trans = dynamic_cast<PESendPayload*>(&trans);
    PESendPayload* chip_trans1 = new PESendPayload();
    if (!orig_trans) {
        wait(IF_DELAY);
        log_error(std::to_string(chiplet_id) + " Invalid transaction type (not PESendPayload)." );
        trans.set_response_status(TLM_GENERIC_ERROR_RESPONSE);
        return;
    }
    
    //manually copy the transaction payload for future use
    sc_time sent_time = orig_trans->timestamp; // Log the time when the core sent the transaction
    chip_trans1->update_timestamp(sc_time_stamp());
    chip_trans1->data = orig_trans->data;
    chip_trans1->src_pe = orig_trans->src_pe;
    chip_trans1->src_chiplet = orig_trans->src_chiplet;
    chip_trans1->dest_pe = orig_trans->dest_pe;
    chip_trans1->dest_chiplet = orig_trans->dest_chiplet;
    chip_trans1->type_flag = orig_trans->type_flag;
    chip_trans1->set_address(orig_trans->get_address());
    chip_trans1->set_command(orig_trans->get_command());
    chip_trans1->set_data_ptr(reinterpret_cast<unsigned char*>(&chip_trans1->data));
    chip_trans1->set_data_length(orig_trans->get_data_length());
    chip_trans1->set_streaming_width(orig_trans->get_streaming_width());
    chip_trans1->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
    chip_trans1->set_dmi_allowed(orig_trans->is_dmi_allowed());

    int address = chip_trans1->get_address(); // address is used to determine wavelength
    // Single-chiplet communication (λM)
    if ((address >= 0x100 && address < 0x200) && (chip_trans1->type_flag == 1)) {
        // Single-chiplet communication
        int chiplet_index = address - 0x100;
        int chiplet_id = chiplet_index - PE_PER_CHIPLET;
        int single_wavelength = chiplet_index;
        
        trans.set_response_status(TLM_OK_RESPONSE);  // Set response status
        log_info(std::to_string(chiplet_id) + " Received Single-chiplet communication, wavelength: " 
                + std::to_string(single_wavelength) + " for chiplet id: " + std::to_string(chiplet_id));
        wait(IF_DELAY);
        log_info(std::to_string(chiplet_id) + " Received input feature: " + std::to_string(chip_trans1->data->id));
        sc_time local_delay_ci_single = SC_ZERO_TIME;  
        // Forward the transaction 
        forward_SingleChiplet_to_O_E_converters(*chip_trans1, local_delay_ci_single, single_wavelength);

    } else {
        log_error(std::to_string(chiplet_id) + " Invalid or unrecognized address: " + std::to_string(address));
    }

    //delete chip_trans1;  // Clean up dynamically allocated memory
    
}

void ChipletInterface::forward_SingleChiplet_to_O_E_converters(PESendPayload& trans, sc_time& delay, int single_wavelength){
    //std::cout << "check: The delay before transmitting to oec: " << delay << std::endl;
    for (int i = 0; i < num_pes; ++i) {
        log_info(std::to_string(chiplet_id) + " Forwarding Single Chiplet transaction: " 
                    + std::to_string(trans.data->id) + " with Wavelength: " + std::to_string(single_wavelength)
                    + " to Optical to Electrical Converter: " + std::to_string(i));
        oec_SingleChiplet_send_sockets[i] -> b_transport(trans, delay); // Send transaction to OEC    
    }    
}


// Handle incoming transactions - Cross chiplet transaction from the Interposer Interface
void ChipletInterface::ii_CrossChiplet_receive_b_transport(tlm_generic_payload& trans, sc_time& delay) {
    PESendPayload* orig_trans = dynamic_cast<PESendPayload*>(&trans);
    PESendPayload* chip_trans2 = new PESendPayload();
    if (!orig_trans) {
        wait(WK_DELAY);
        log_error( std::to_string(chiplet_id) + " Invalid transaction type (not PESendPayload)." );
        trans.set_response_status(TLM_GENERIC_ERROR_RESPONSE);
        return;
    }

    sc_time current_time = sc_time_stamp();
    //manually copy the transaction payload for future use
    sc_time sent_time = orig_trans->timestamp; // Log the time when the core sent the transaction
    chip_trans2->data = orig_trans->data;
    chip_trans2->src_pe = orig_trans->src_pe;
    chip_trans2->src_chiplet = orig_trans->src_chiplet;
    chip_trans2->dest_pe = orig_trans->dest_pe;
    chip_trans2->dest_chiplet = orig_trans->dest_chiplet;
    chip_trans2->type_flag = orig_trans->type_flag;
    chip_trans2->update_timestamp(current_time);
    chip_trans2->set_address(orig_trans->get_address());
    chip_trans2->set_command(orig_trans->get_command());
    chip_trans2->set_data_ptr(reinterpret_cast<unsigned char*>(&chip_trans2->data));
    chip_trans2->set_data_length(orig_trans->get_data_length());
    chip_trans2->set_streaming_width(orig_trans->get_streaming_width());
    chip_trans2->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
    chip_trans2->set_dmi_allowed(orig_trans->is_dmi_allowed());

    int address = chip_trans2->get_address(); // address is used to determine wavelength
    if ((address >= 0x200)) {        
        int cross_wavelength = address - 0x200;
        log_info(std::to_string(chiplet_id) + " Received Cross-chiplet communication: "  
                + std::to_string(chip_trans2->data->id) + ", with wavelength: " + std::to_string(cross_wavelength));
        wait(WK_DELAY);
        //std::cout << "check: The corss wavelength at chiplet is " << cross_wavelength << std::endl;
        // Forward the transaction to the chiplet interface
        sc_time local_delay_ci_cross = SC_ZERO_TIME;
        forward_CrossChiplet_to_O_E_converters(*chip_trans2, local_delay_ci_cross, cross_wavelength);
       
    } else {
        log_error(std::to_string(chiplet_id) + " Invalid or unrecognized address: " + std::to_string(address));
    }

    //delete chip_trans2;  // Clean up dynamically allocated memory
    trans.set_response_status(TLM_OK_RESPONSE);  // Set response status
}

void ChipletInterface::forward_CrossChiplet_to_O_E_converters(PESendPayload& trans, sc_time& delay, int cross_wavelength){
    //std::cout << "check: The delay before transmitting cross chiplet data to oec: " << delay << std::endl;
    oec_CrossChiplet_send_sockets[cross_wavelength] -> b_transport(trans, delay); // Send transaction to OEC
    log_info(std::to_string(chiplet_id) + " Forwarding Cross chiplet transaction: " + 
            std::to_string(trans.data->id) + " with Wavelength: " + std::to_string(cross_wavelength) 
            + " to Optical to Electrical Converter: " + std::to_string(cross_wavelength));  
}

//Handle incoming transactions from the OEC OF/PE_TO_PE
void ChipletInterface::oec_b_transport(tlm_generic_payload& trans, sc_time& delay) {
    PESendPayload* orig_trans = dynamic_cast<PESendPayload*>(&trans);
    PESendPayload* rx_trans2 = new PESendPayload();
    if (!orig_trans) {
        wait(OF_DELAY);
        log_error(std::to_string(chiplet_id) + "  Invalid transaction type (not PESendPayload).");
        trans.set_response_status(TLM_GENERIC_ERROR_RESPONSE);
        return;
    }
    sc_time current_time = sc_time_stamp();
    //manually copy the transaction payload for future use
    sc_time sent_time = orig_trans->timestamp; // Log the time when the core sent the transaction
    rx_trans2->data = orig_trans->data;
    rx_trans2->src_chiplet = orig_trans->src_chiplet;
    rx_trans2->src_pe = orig_trans->src_pe;
    rx_trans2->src_pe = orig_trans->src_pe;
    rx_trans2->src_chiplet = orig_trans->src_chiplet;
    rx_trans2->dest_pe = orig_trans->dest_pe;
    rx_trans2->dest_chiplet = orig_trans->dest_chiplet;
    rx_trans2->type_flag = orig_trans->type_flag;
    rx_trans2->update_timestamp(current_time);
    rx_trans2->set_address(orig_trans->get_address());
    rx_trans2->set_command(orig_trans->get_command());
    rx_trans2->set_data_ptr(reinterpret_cast<unsigned char*>(&rx_trans2->data));
    rx_trans2->set_data_length(orig_trans->get_data_length());
    rx_trans2->set_streaming_width(orig_trans->get_streaming_width());
    rx_trans2->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
    rx_trans2->set_dmi_allowed(orig_trans->is_dmi_allowed());

    int address = rx_trans2->get_address(); // address is used to determine wavelength
    int chiplet_index = address - 0x300;
    int unicast_wavelength = chiplet_index;

    wait(OF_DELAY);
    log_info(std::to_string(chiplet_id) + " Chiplet Interface received Packet from PE: " 
            + std::to_string(rx_trans2->src_pe) + " through OEC with data: " 
            + std::to_string(rx_trans2->data->id));
    trans.set_response_status(TLM_OK_RESPONSE);  // Set response status
    // Forward the transaction to the Interposer interface 
    sc_time local_delay_ci_OF = SC_ZERO_TIME;
    forward_data_to_interposer(*rx_trans2, local_delay_ci_OF);    
    //delete rx_trans2;  // Clean up dynamically allocated memory   
}

void ChipletInterface::forward_data_to_interposer(PESendPayload& trans, sc_time& delay){
    //std::cout << "check: The delay before transmitting OF to II: " << delay << std::endl;
    log_info(std::to_string(chiplet_id) + " Forwarding the data " + std::to_string(trans.data->id) 
            + " from PE: " + std::to_string(trans.src_pe) + " to Interposer interface: " 
            + std::to_string(trans.src_chiplet));
    interposer_send_socket->b_transport(trans, delay); 
}

//The logging for chiplet interface is done using C_LOG defined in noc_logger
void ChipletInterface::log_info(std::string msg){
    SC_REPORT_INFO(C_LOG, msg.c_str());
}

void ChipletInterface::log_error(std::string msg){
    SC_REPORT_ERROR(C_LOG, msg.c_str());
}