#include "Global_Buffer.h"

// Constructor implementation
GlobalBuffer::GlobalBuffer(sc_module_name name, int gb_id, int num_chiplets, int num_pes)
    : sc_module(name),
      gb_id(gb_id),
      num_chiplets(num_chiplets),
      num_pes(num_pes), 
      SingleChiplet_fifo(SINGLE_CHIPLET_FIFO_CAPACITY),
      CrossChiplet_fifo(CROSS_CHIPLET_FIFO_CAPACITY),
      OF_fifo(OF_FIFO_CAPACITY)
    {
    
    sc_report_handler::set_actions(GB_LOG, SC_INFO, SC_LOG|SC_DISPLAY);
    for (int i = 0; i < num_chiplets; ++i) {
        // Register b_transport callback for each target socket to receive OF or data packets from PEs
        receive_to_GB_sockets[i].register_b_transport(this, &GlobalBuffer::receive_to_gb_b_transport);
    }
    //register non-blocking transport to receive data from NI of GB
    ni_to_gb_receive_socket.register_nb_transport_fw(this, &GlobalBuffer::gb_nb_transport_fw);

    // SC_THREAD for processing SingleChiplet, CrossChiplets, and data reception
    SC_THREAD(single_chiplet_communication);

    SC_THREAD(cross_chiplet_communication);

    SC_THREAD(OF_communication);
}

tlm::tlm_sync_enum GlobalBuffer::gb_nb_transport_fw(tlm::tlm_generic_payload& trans, tlm::tlm_phase& phase, sc_time& delay) {
    if (phase == tlm::BEGIN_REQ) {
        sc_time current_time = sc_time_stamp();
        //manually copy the transaction payload for future use
        PESendPayload* orig_trans = dynamic_cast<PESendPayload*>(&trans);
        PESendPayload* gb_trans1 = new PESendPayload(); 
        if (!orig_trans) {
            log_error(" Invalid transaction type (not PESendPayload)." );
            return tlm::TLM_UPDATED;
        }        
        sc_time sent_time = orig_trans->timestamp; 
        gb_trans1->data = orig_trans->data;
        gb_trans1->src_pe = orig_trans->src_pe;
        gb_trans1->src_chiplet = orig_trans->src_chiplet;
        gb_trans1->dest_pe = orig_trans->dest_pe;
        gb_trans1->dest_chiplet = orig_trans->dest_chiplet;
        gb_trans1->type_flag = orig_trans->type_flag;
        int dst_pe = gb_trans1->dest_pe;
        int type_flag = gb_trans1->type_flag;
        int dst_chiplet_id = orig_trans->dest_chiplet;
        gb_trans1->update_timestamp(current_time);  
        gb_trans1->set_command(orig_trans->get_command());
        gb_trans1->set_data_ptr(reinterpret_cast<unsigned char*>(&gb_trans1->data));
        gb_trans1->set_data_length(orig_trans->get_data_length());
        gb_trans1->set_streaming_width(orig_trans->get_streaming_width());
        gb_trans1->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
        gb_trans1->set_dmi_allowed(orig_trans->is_dmi_allowed());
        gb_trans1->set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
        int address = gb_trans1->get_address(); // address is used to determine wavelength

        //std::cout << "check-GB: flag of packet with id "<< gb_trans1->data->id <<" is "<< type_flag <<std::endl;
        
        if(type_flag == 2){
            if (CrossChiplet_fifo.num_free() == 0) {
                log_info("Cross Chiplet FIFO full, cannot accept transaction with data: " 
                        + std::to_string(gb_trans1->data->id));
                delay = sc_time(RETRY_DELAY); // Delay before retry 
                return tlm::TLM_UPDATED; // Indicate to retry later
            } else {
                log_info("Received transaction is a weight kernal with id: "  
                    + std::to_string(gb_trans1->data->id) + " from Network Interface of GB");
                CrossChiplet_fifo.write(gb_trans1); // Push the transaction into the input FIFO by blocking write
            }
        } else if(type_flag == 1){
            if (SingleChiplet_fifo.num_free() == 0) {
                log_info("Single Chiplet FIFO full, cannot accept transaction with data: " 
                        + std::to_string(gb_trans1->data->id));
                delay = sc_time(RETRY_DELAY); // Delay before retry 
                return tlm::TLM_UPDATED; // Indicate to retry later
            } else {
               
                log_info("Received transaction is an input feature with id: " 
                        + std::to_string(gb_trans1->data->id) + " from Network Interface of GB");

                SingleChiplet_fifo.write(gb_trans1); // Push the transaction into the input FIFO by blocking write
            }
        }
        
        phase = tlm::END_REQ;  // Mark the end of the request phase
        delay = sc_time(UPDATE_DELAY);  // Add some delay
        return tlm::TLM_UPDATED;
    }
    return tlm::TLM_ACCEPTED;
}


// Process to receive Packets from all PEs
void GlobalBuffer::receive_to_gb_b_transport(tlm_generic_payload& trans, sc_time& delay) {
    PESendPayload* orig_trans = dynamic_cast<PESendPayload*>(&trans);
    if (!orig_trans) {
        log_error("  Invalid transaction type (not PESendPayload).");
        trans.set_response_status(TLM_GENERIC_ERROR_RESPONSE);
        return;
    }
    PESendPayload* rx_trans4 = new PESendPayload();//mm.allocate();
    rx_trans4->acquire();
    //manually copy the transaction payload for future use
    sc_time current_time = sc_time_stamp();
    sc_time sent_time = orig_trans->timestamp; // Log the time when the core sent the transaction
    rx_trans4->data = orig_trans->data;
    rx_trans4->src_chiplet = orig_trans->src_chiplet;
    int chiplet_id = rx_trans4->src_chiplet;
    rx_trans4->src_pe = orig_trans->src_pe;
    int pe_id = rx_trans4->src_pe;
    rx_trans4->dest_pe = orig_trans->dest_pe;
    int dest_pe = rx_trans4->dest_pe;
    rx_trans4->dest_chiplet = orig_trans->dest_chiplet;
    int dest_chiplet = rx_trans4->dest_chiplet;
    rx_trans4->update_timestamp(current_time);
    rx_trans4->type_flag = orig_trans->type_flag;
    int type_flag = rx_trans4->type_flag;
    rx_trans4->set_command(orig_trans->get_command());
    rx_trans4->set_data_ptr(reinterpret_cast<unsigned char*>(&rx_trans4->data));
    rx_trans4->set_data_length(orig_trans->get_data_length());
    rx_trans4->set_streaming_width(orig_trans->get_streaming_width());
    rx_trans4->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
    rx_trans4->set_dmi_allowed(orig_trans->is_dmi_allowed());

    int address = orig_trans->get_address(); // address is used to determine wavelength
    //std::cout << " chack: The old address of incoming communication is " << address << std::endl;
    int chiplet_index = address - 0x300;
    int unicast_wavelength = chiplet_index;
    
    //std::cout << " check: The new address for cross chiplet communication is " << rx_trans4->get_address() << std::endl;
    wait(OF_DELAY);
    //std::cout << "The delay when received OF in GB: " << delay << std::endl;
    log_info(" Received transaction from PE: " + std::to_string(rx_trans4->src_pe) + " from chiplet: "
            + std::to_string(rx_trans4->src_chiplet) + " through Interposer Interface with data: " 
            + std::to_string(rx_trans4->data->id));
       
    //update_counters(chiplet_id, pe_id);
    if(type_flag == 0) {
        log_info("Received transaction: " + std::to_string(rx_trans4->data->id) + 
                " is a packet from the source PE: " + std::to_string(pe_id) + " for destination PE: "
                + std::to_string(dest_pe));
            
        CrossChiplet_fifo.write(rx_trans4);
        log_info("Written the packet data id: " + std::to_string(rx_trans4->data->id) + " From the PE: " + 
                std::to_string(pe_id) + " to the crosschiplet fifo");
        trans.set_response_status(TLM_OK_RESPONSE);  // Set response status
    } else if(type_flag == 3) {
        log_info("Received transaction is an Output Feature (OF): " + std::to_string(rx_trans4->data->id) 
                + " from PE: " + std::to_string(pe_id));
        OF_fifo.write(rx_trans4);
        log_info("Written the Output Feature (OF) data id: " + std::to_string(rx_trans4->data->id) + " from the PE: " + 
                std::to_string(pe_id) + " to the OF fifo");
    }

    //delete rx_trans4;  // Clean up dynamically allocated memory    
}

void GlobalBuffer::cross_chiplet_communication() {
    while(true) {
        PESendPayload* gb_trans2 = new PESendPayload();//mm.allocate(); 
        gb_trans2->acquire();
        PESendPayload* fifo_trans1 = CrossChiplet_fifo.read();
        //std::cout << "check: the id of data read from cross chiplet fifo is: " << fifo_trans1->data->id << std::endl;
        //manually copy the transaction payload for future use
        sc_time current_time = sc_time_stamp();
        sc_time sent_time = fifo_trans1->timestamp; // Log the time when the core sent the transaction
        gb_trans2->data = fifo_trans1->data;
        gb_trans2->src_chiplet = fifo_trans1->src_chiplet;
        gb_trans2->src_pe = fifo_trans1->src_pe;
        gb_trans2->dest_pe = fifo_trans1->dest_pe;
        int dest_pe = gb_trans2->dest_pe;
        gb_trans2->dest_chiplet = fifo_trans1->dest_chiplet;
        int dest_chiplet = gb_trans2->dest_chiplet;
        //DISPLAY_PARAMETERS("destination chiplet read from cross chiplet fifo", dest_chiplet);
        gb_trans2->update_timestamp(current_time);
        gb_trans2->type_flag = fifo_trans1->type_flag;
        //gb_trans2->set_address(fifo_trans1->get_address());
        gb_trans2->set_command(fifo_trans1->get_command());
        gb_trans2->set_data_ptr(reinterpret_cast<unsigned char*>(&gb_trans2->data));
        gb_trans2->set_data_length(fifo_trans1->get_data_length());
        gb_trans2->set_streaming_width(fifo_trans1->get_streaming_width());
        gb_trans2->set_byte_enable_ptr(fifo_trans1->get_byte_enable_ptr());
        gb_trans2->set_dmi_allowed(fifo_trans1->is_dmi_allowed());
        
        int type_flag = gb_trans2->type_flag;
        //std::cout << "check- GB: The type flag at the cross chiplet fifo is: " << type_flag << std::endl;
        sc_core::sc_event trigger_events[CHIPLET_NO];
        sc_core::sc_time local_delay_gb_cross = SC_ZERO_TIME;
        if(type_flag == 2) {
            for (int chiplet_id = 0; chiplet_id < CHIPLET_NO; ++chiplet_id) {
                int pe_index = node_to_pe(dest_pe);
                int wk_address = pe_index + 0x200;
                gb_trans2->set_address(wk_address);
                //std::cout << "check: the address at global buffer for cross chiplet transaction is: " << wk_address << std::endl;
                log_info("Cross chiplet transaction of weight kernal " 
                        + std::to_string(gb_trans2->data->id) + " to interposer Interface " + std::to_string(chiplet_id));
                sc_core::sc_spawn([&, chiplet_id, dest_pe, gb_trans2]() {
                    CrossChiplet_from_GB_sockets[chiplet_id]->b_transport(*gb_trans2, local_delay_gb_cross);
                    if(gb_trans2->get_response_status() == TLM_OK_RESPONSE) {
                        log_info("Transaction of Cross chiplet transaction of weight kernal " + std::to_string(gb_trans2->data->id) 
                                + " completed successfully from Global Buffer to PE");
                    } else {
                        log_error ("Error in transmitting Cross chiplet transaction: " 
                                + std::to_string(gb_trans2->data->id) + " from Global Buffer to PE");
                    }
                 });
            } 
        } else if(type_flag == 0) {
            for (int chiplet_id = 0; chiplet_id < CHIPLET_NO; ++chiplet_id) {                   
                //The addressing scheme based on the assemption that the number of chiplets and the number of PEs are equal, otherwise the PE_OFFSET value needed to deducted from dest_pe
                int pe_wavelength = node_to_pe(dest_pe);
                //std::cout << "check- GB: The wavelength for cross chiplet transaction from destination PE " << pe_wavelength << std::endl;
                int new_address = 0x200 + pe_wavelength;
                //DISPLAY_PARAMETERS("new_address", new_address);
                gb_trans2->set_address(new_address);        
                log_info("Transmitting packet for PE: " + std::to_string(dest_pe) + " as cross chiplet transaction with data id: " 
                        + std::to_string(gb_trans2->data->id) + " to interposer Interface " + std::to_string(chiplet_id));
                //sc_core::sc_spawn_options opt;
                //std::string process_name = "cross_chiplet_spawn_" + std::to_string(chiplet_id);
                //opt.set_sensitivity(&trigger_events[chiplet_id]); 
                //opt.dont_initialize();
                sc_core::sc_spawn([&, chiplet_id, dest_pe, gb_trans2]() {
                    
                    CrossChiplet_from_GB_sockets[chiplet_id]->b_transport(*gb_trans2, local_delay_gb_cross);
                    
                    if(gb_trans2->get_response_status() == TLM_OK_RESPONSE) {
                        log_info("Transaction of Cross chiplet transaction of data packet " + std::to_string(gb_trans2->data->id) 
                                + " completed successfully from Global Buffer to PE");
                    } else {
                        log_error ("Error in transmitting Cross chiplet transaction: " 
                                + std::to_string(gb_trans2->data->id) + " from Global Buffer to PE");
                    }
                        
                });   

            }
        }    
    }
}

void GlobalBuffer::single_chiplet_communication() {  
    while(true) {
       
        PESendPayload* gb_trans3 = new PESendPayload();//mm.allocate(); 
        gb_trans3->acquire();
        PESendPayload* fifo_trans2 = SingleChiplet_fifo.read();

        //manually copy the transaction payload for future use
        sc_time current_time = sc_time_stamp();
        sc_time sent_time = fifo_trans2->timestamp; // Log the time when the core sent the transaction
        gb_trans3->data = fifo_trans2->data;
        gb_trans3->src_chiplet = fifo_trans2->src_chiplet;
        gb_trans3->src_pe = fifo_trans2->src_pe;
        gb_trans3->dest_pe = fifo_trans2->dest_pe;
        gb_trans3->dest_chiplet = fifo_trans2->dest_chiplet;
        gb_trans3->update_timestamp(current_time);
        gb_trans3->type_flag = fifo_trans2->type_flag;
        gb_trans3->set_command(fifo_trans2->get_command());
        gb_trans3->set_data_ptr(reinterpret_cast<unsigned char*>(&gb_trans3->data));
        gb_trans3->set_data_length(fifo_trans2->get_data_length());
        gb_trans3->set_streaming_width(fifo_trans2->get_streaming_width());
        gb_trans3->set_byte_enable_ptr(fifo_trans2->get_byte_enable_ptr());
        gb_trans3->set_dmi_allowed(fifo_trans2->is_dmi_allowed());
        //std::cout << "check: The single chiplet transaction in fifo is " << gb_trans3->data->id << std::endl;

        int dest_pe = gb_trans3->dest_pe;
        int dest_chiplet = gb_trans3->dest_chiplet;
        int chiplet_index = 0x100 + (PE_PER_CHIPLET + dest_chiplet);
        gb_trans3->set_address(chiplet_index);
        //FUNCTION_INFO("Single Chiplet comm address, for data id ", chiplet_index, gb_trans3->data->id)
        //std::cout << "check: The address for single chiplet communication is " << chiplet_index << std::endl;
        sc_time delay = SC_ZERO_TIME;
        log_info("Transmitting Input features as Single chiplet transaction: " 
                + std::to_string(gb_trans3->data->id) + " to interposer Interface " + std::to_string(dest_chiplet));

        SingleChiplet_from_GB_sockets[dest_chiplet]->b_transport(*gb_trans3, delay);

        if(gb_trans3->get_response_status() == TLM_OK_RESPONSE) {
            log_info("Transaction of Single chiplet transaction " + std::to_string(gb_trans3->data->id) 
                    + " completed successfully from Global Buffer to PE");
        } else {
            log_error ("Error in transmitting Single chiplet transaction: " 
                    + std::to_string(gb_trans3->data->id) + " from Global Buffer to Interposer Interface");
        }
          	
    }
}

void GlobalBuffer::OF_communication() {
    PESendPayload* payload;
    while(true) {
        payload = OF_fifo.read();
        int src_pe = payload->src_pe;
        int dest_chiplet = payload->dest_chiplet;
        tlm::tlm_phase phase = tlm::BEGIN_REQ;
        sc_time delay = SC_ZERO_TIME;
        tlm::tlm_sync_enum status = gb_to_ni_send_socket->nb_transport_fw(*payload, phase, delay);
        log_info("Transmitting Output Feature (OF): " + std::to_string(payload->data->id) + " received from PE: " + std::to_string(src_pe) 
                    + " to network Interface ");
        if (status == tlm::TLM_COMPLETED) {
            log_info("Completed transaction, data ID: " + std::to_string(payload->data->id));   
        } else if (status == tlm::TLM_ACCEPTED) {
            log_info("Accepted transaction, waiting for completion.");
        } else if (status == tlm::TLM_UPDATED) {
            log_info("Updated transaction, waiting for completion.");
        } else {
            log_error("Unexpected response from Network Interface ");
            //delete payload;
        } 
    }
}

void GlobalBuffer::log_info(std::string msg){
    SC_REPORT_INFO(GB_LOG, msg.c_str());
}

void GlobalBuffer::log_error(std::string msg){
    SC_REPORT_ERROR(GB_LOG, msg.c_str());
}