#include "Interposer_Interface.h"
#include <iostream>

// Constructor implementation
InterposerInterface::InterposerInterface(sc_module_name name)
    : sc_module(name)
{
    sc_report_handler::set_actions(I_LOG, SC_INFO, SC_LOG|SC_DISPLAY);
    // Register b_transport methods for GB and Chiplet interfaces
    gb_SingleChiplet_receive_socket.register_b_transport(this, &InterposerInterface::gb_SingleChiplet_b_transport);
    gb_CrossChiplet_receive_socket.register_b_transport(this, &InterposerInterface::gb_CrossChiplet_b_transport);
    chiplet_receive_socket.register_b_transport(this, &InterposerInterface::chiplet_b_transport);
    
}

// Handle incoming transactions from the Global Buffer for single chiplet broadcast
void InterposerInterface::gb_SingleChiplet_b_transport(tlm_generic_payload& trans, sc_time& delay) {
    PESendPayload* orig_trans = dynamic_cast<PESendPayload*>(&trans);
    if (!orig_trans) {
        wait(IF_DELAY);
        log_error("Invalid transaction type (not PESendPayload).");
        trans.set_response_status(TLM_GENERIC_ERROR_RESPONSE);
        return;
    }
    // Copy payload manually
    PESendPayload* ii_trans1 = new PESendPayload();
    sc_time current_time = sc_time_stamp();
    sc_time sent_time = orig_trans->timestamp; // Log the time when the core sent the transaction
    ii_trans1->data = orig_trans->data;
    ii_trans1->src_pe = orig_trans->src_pe;
    ii_trans1->src_chiplet = orig_trans->src_chiplet;
    ii_trans1->dest_pe = orig_trans->dest_pe;
    ii_trans1->dest_chiplet = orig_trans->dest_chiplet;
    ii_trans1->type_flag = orig_trans->type_flag;
    ii_trans1->update_timestamp(current_time);
    ii_trans1->set_address(orig_trans->get_address());
    ii_trans1->set_command(orig_trans->get_command());
    ii_trans1->set_data_ptr(reinterpret_cast<unsigned char*>(&ii_trans1->data));
    ii_trans1->set_data_length(orig_trans->get_data_length());
    ii_trans1->set_streaming_width(orig_trans->get_streaming_width());
    ii_trans1->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
    ii_trans1->set_dmi_allowed(orig_trans->is_dmi_allowed());
    
    int address = ii_trans1->get_address(); // address is used to determine wavelength
    if (address >= 0x100 && address < 0x200) {
        // Single-chiplet communication
        int chiplet_index = address - 0x100;
        int chiplet_id = chiplet_index - PE_PER_CHIPLET;  
        int single_wavelength = chiplet_index;
        trans.set_response_status(TLM_OK_RESPONSE);  // Set response status
        wait(IF_DELAY);
        log_info(std::to_string(interposer_id) + " Received input feature: " + std::to_string(ii_trans1->data->id) + " with Wavelength: " + std::to_string(single_wavelength));
        sc_time local_delay_ii = SC_ZERO_TIME;
        // Forward the transaction to the chiplet interface
        chiplet_SingleChiplet_send_transaction(*ii_trans1, local_delay_ii, single_wavelength);
        
    } else {
        log_error(std::to_string(interposer_id) + " Invalid or unrecognized address: " + std::to_string(address));
    }

    //delete ii_trans1;  // Clean up dynamically allocated memory
    
}

void InterposerInterface::chiplet_SingleChiplet_send_transaction(PESendPayload& trans, sc_time& delay, int single_wavelength) {
    // Forward the transaction to the chiplet interface
    //std::cout << "check: The delay before transmitting single chiplet data to CI: " << delay << std::endl;
    log_info(std::to_string(interposer_id) + " Forwarding Single Chiplet transaction: " 
            + std::to_string(trans.data->id) + " with Wavelength: " + std::to_string(single_wavelength) 
            + " to chiplet interface");
    chiplet_SingleChiplet_send_socket->b_transport(trans, delay);  // Send transaction to Chiplet Interface  
}

// Handle incoming Cross chiplet transaction from the Global Buffer
void InterposerInterface::gb_CrossChiplet_b_transport(tlm_generic_payload& trans, sc_time& delay) {
    PESendPayload* orig_trans = dynamic_cast<PESendPayload*>(&trans);
    if (!orig_trans) {
        wait(WK_DELAY);
        log_error(" Invalid transaction type (not PESendPayload)");
        trans.set_response_status(TLM_GENERIC_ERROR_RESPONSE);
        return;
    }

    PESendPayload* ii_trans1 = new PESendPayload();
    sc_time current_time = sc_time_stamp();
    //manually copy the transaction payload for future use
    ii_trans1->acquire();
    sc_time sent_time = orig_trans->timestamp; // Log the time when the core sent the transaction
    ii_trans1->data = orig_trans->data;
    ii_trans1->src_pe = orig_trans->src_pe;
    ii_trans1->src_chiplet = orig_trans->src_chiplet;
    ii_trans1->dest_pe = orig_trans->dest_pe;
    ii_trans1->dest_chiplet = orig_trans->dest_chiplet;
    ii_trans1->type_flag = orig_trans->type_flag;
    ii_trans1->update_timestamp(current_time);
    ii_trans1->set_address(orig_trans->get_address());
    ii_trans1->set_command(orig_trans->get_command());
    ii_trans1->set_data_ptr(reinterpret_cast<unsigned char*>(&ii_trans1->data));
    ii_trans1->set_data_length(orig_trans->get_data_length());
    ii_trans1->set_streaming_width(orig_trans->get_streaming_width());
    ii_trans1->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
    ii_trans1->set_dmi_allowed(orig_trans->is_dmi_allowed());

    int address = ii_trans1->get_address(); // address is used to determine wavelength
   // std::cout << "check: the address at interposer interface for cross chiplet transaction is: " << address << std::endl;
    // Cross-chiplet communication (λ0, λ1, λ2, ...)
    if (address >= 0x200) {
        int cross_wavelength = address - 0x200;
        //std::cout << "check: the wavelength at interposer interface for cross chiplet transaction is: " << cross_wavelength << std::endl;
        sc_time path_delay = WK_DELAY * interposer_id;
        wait(path_delay);
        double pd = path_delay.to_double();
        FUNCTION_INFO("Delay at Interposer", (pd) );
        log_info(std::to_string(interposer_id) + " Received cross-chiplet data: " + std::to_string(ii_trans1->data->id) 
                + " with Wavelength: " + std::to_string(cross_wavelength));
        trans.set_response_status(TLM_OK_RESPONSE);        
        log_info(std::to_string(interposer_id) + " Splitting data for PE: " + std::to_string(cross_wavelength) 
                + " to chiplet:" + std::to_string(interposer_id));
        
        // Forward modified data to chiplet interface
        sc_time local_delay_ii_cross = SC_ZERO_TIME;
        chiplet_CrossChiplet_send_transaction(*ii_trans1, local_delay_ii_cross, cross_wavelength);
    } else {
        log_error(std::to_string(interposer_id) + " Invalid or unrecognized address: " + std::to_string(address));    
    }

    //delete ii_trans1;  // Clean up dynamically allocated memory
}

void InterposerInterface::chiplet_CrossChiplet_send_transaction(PESendPayload& trans, sc_time& delay, int cross_wavelength){
    //std::cout << "check: The delay before transmitting cross chiplet data to II: " << delay << std::endl;
    chiplet_CrossChiplet_send_socket->b_transport(trans, delay);  // Send transaction to Chiplet Interface
    log_info(std::to_string(interposer_id) + " Forwarding Cross chiplet transaction: "
             + std::to_string(trans.data->id) + " with Wavelength: " + std::to_string(cross_wavelength) 
             + " to chiplet interface");
}

//Handle incoming transactions from the Chiplet Interface OF or PE_TO_PE
void InterposerInterface::chiplet_b_transport(tlm_generic_payload& trans, sc_time& delay) {
    PESendPayload* orig_trans = dynamic_cast<PESendPayload*>(&trans);
    if (!orig_trans) {
        wait(OF_DELAY);
        log_error(std::to_string(interposer_id) + "  Invalid transaction type (not PESendPayload).");
        trans.set_response_status(TLM_GENERIC_ERROR_RESPONSE);
        return;
    }

    PESendPayload* rx_trans3 = new PESendPayload();
    sc_time current_time = sc_time_stamp();
    //manually copy the transaction payload for future use
    sc_time sent_time = orig_trans->timestamp;
    rx_trans3->data = orig_trans->data;
    rx_trans3->src_pe = orig_trans->src_pe;
    rx_trans3->src_chiplet = orig_trans->src_chiplet;
    rx_trans3->dest_pe = orig_trans->dest_pe;
    rx_trans3->dest_chiplet = orig_trans->dest_chiplet;
    rx_trans3->type_flag = orig_trans->type_flag;
    rx_trans3->update_timestamp(current_time);
    rx_trans3->set_address(orig_trans->get_address());
    rx_trans3->set_command(orig_trans->get_command());
    rx_trans3->set_data_ptr(reinterpret_cast<unsigned char*>(&rx_trans3->data));
    rx_trans3->set_data_length(orig_trans->get_data_length());
    rx_trans3->set_streaming_width(orig_trans->get_streaming_width());
    rx_trans3->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
    rx_trans3->set_dmi_allowed(orig_trans->is_dmi_allowed());

    int address = rx_trans3->get_address(); // address is used to determine wavelength

    int chiplet_index = address - 0x300;
   
    int unicast_wavelength = chiplet_index;

    wait(OF_DELAY);

    log_info(std::to_string(interposer_id) + " received data from PE: " + std::to_string(rx_trans3->src_pe) 
            + " through OEC with data: " + std::to_string(rx_trans3->data->id));
    trans.set_response_status(TLM_OK_RESPONSE);  // Set response status

    // Forward the transaction to the Global Buffer 
    sc_time local_delay_ii_OF = SC_ZERO_TIME;
    gb_send_transaction(*rx_trans3, local_delay_ii_OF);

    //delete rx_trans3;  // Clean up dynamically allocated memory   
}

void InterposerInterface::gb_send_transaction(PESendPayload& trans, sc_time& delay){
    //std::cout << "check: The delay before transmitting OF to GB: " << delay << std::endl;
    gb_send_socket->b_transport(trans, delay);    
    log_info(std::to_string(interposer_id) + " Forwarding the datas from PE: " + std::to_string(trans.src_pe) 
            + " to Global Buffer ");  
}

void InterposerInterface::log_info(std::string msg){
    SC_REPORT_INFO(I_LOG, msg.c_str());
}

void InterposerInterface::log_error(std::string msg){
    SC_REPORT_ERROR(I_LOG, msg.c_str());
}