#include "NetworkInterfaceGB.h"
#include <iostream>
#include "ratatoskrUtils/traffic/Flit.h"

using namespace std;

NetworkInterfaceGB::NetworkInterfaceGB(sc_module_name nm, Node& node, TokenManager& token_ref) : 
    NetworkInterface(nm, node, token_ref), lastReceivedCreditID(-1), credit_counter(0), resp_in_progress(false)
        {
            setup_logger();
            sc_report_handler::set_actions(NI_GB_LOG, SC_INFO, SC_LOG|SC_DISPLAY);
            ni_receive_socket.register_nb_transport_fw(this, &NetworkInterfaceGB::ni_nb_transport_fw);
    try {
        this->id = ni_gb_id;
        this->dbid = rep.registerElement("NetworkInterfaceGB", this->id);
        this->node = node;
        this->packetPortContainer = new PacketPortContainer(
                ("NI_PACKET_CONTAINER" + std::to_string(this->id)).c_str());
    }
    catch (exception& e) {
        log_error("Initialization Error: " + string(e.what()));
    }

    SC_THREAD(thread);
    sensitive << clk.pos() << clk.neg();

    SC_METHOD(receivePacketFromPE);
    sensitive << packetPortContainer->portValidIn.pos();

}

NetworkInterfaceGB::~NetworkInterfaceGB() {
    delete packetPortContainer;
    
}

void NetworkInterfaceGB::initialize(){
    /* NOT IMPLEMENTED */
}

void NetworkInterfaceGB::bind(Connection* con, SignalContainer* sigContIn, SignalContainer* sigContOut) {
    if (con == nullptr)
        packetPortContainer->bind(sigContIn, sigContOut);
}

void NetworkInterfaceGB::receiveFlitFromRouter() {
    /* NOT IMPLEMENTED */
}

void NetworkInterfaceGB::receivePacketFromPE() {
    log_info(std::to_string(ni_gb_id) + " program entered in function receivePacketFromPE");
    if (packetPortContainer->portValidIn.posedge()) {

        //LOG(globalReport.verbose_pe_function_calls, "NI" << this->id << "(Node" << node.id << ")\t\t- receive()");
        Packet* packet = packetPortContainer->portDataIn.read();
        log_info(std::to_string(ni_gb_id) + " packet received in function receivePacketFromPE " + to_string(packet->id));
        generateFlitsForPacket(packet);
        packet_send_queue.push(packet);
    }
}

void NetworkInterfaceGB::getDestination(Packet* p, std::string pos[2]){
    pos[0] = to_string(node.pos.x) + "," +
            to_string(node.pos.y) + "," + 
            to_string(node.pos.z);
    pos[1] = to_string(p->dst.pos.x) + "," +
            to_string(p->dst.pos.y) + "," +
            to_string(p->dst.pos.z);
}

void NetworkInterfaceGB::generateFlitsForPacket(Packet* p) {
    int flitsPerPacket = p->size;
    for (int i = 0; i<flitsPerPacket; ++i) {
        FlitType flitType;
        if (flitsPerPacket==1)
            flitType = SINGLE;
        else if (i%flitsPerPacket==0)
            flitType = HEAD;
        else if (i%flitsPerPacket==flitsPerPacket-1)
            flitType = TAIL;
        else
            flitType = BODY;
        int seqNum = i%flitsPerPacket;
        Flit* current_flit = new Flit(flitType, seqNum, p, p->dataType, sc_time_stamp().to_double());
        p->toTransmit.push_back(current_flit->id);
        p->flits.push_back(current_flit);
        
        std::string start_end_pos[2];
        getDestination(p, start_end_pos);
        log_info(std::to_string(ni_gb_id) + " (of GB): Flits generated for packet of id " + to_string(p->id) 
        + " to be sent to "+ std::to_string(p->dst.id) + " from " + std::to_string(p->src.id));
    }
}

void NetworkInterfaceGB::send_packet_to_gb_from_ni() {
    PESendPayload* trans = nullptr;
        
    sc_time delay = sc_time(0, SC_NS);
    bool retry = false;
    do {
        if (!retry) { //When retry is false new flits are generated from packets
            Packet* p = packet_send_queue.front();
            // if no flits to transmit, generate them
            if (p->toTransmit.empty()){ 
                generateFlitsForPacket(p);
            }
            // get current flit to send
            flitID_t f_id = p->toTransmit.front();
            auto iter = find_if(p->flits.begin(), p->flits.end(),
                    [&f_id](Flit* f) { return f->id==f_id; });
            Flit* current_flit = *iter;
            current_flit->injectionTime = sc_time_stamp().to_double();

            auto toDelete_pos = find(p->toTransmit.begin(), p->toTransmit.end(), current_flit->id);
            p->toTransmit.erase(toDelete_pos);
            p->inTransmit.push_back(current_flit->id);

            if (p->toTransmit.empty()) {
                packet_send_queue.pop();
            }

            unsigned int src_pe = p->src.id;
            //std::cout << "check: The source PE is: " << src_pe << std::endl; 
            unsigned int src_chiplet = (ni_gb_id - 1); 
            //std::cout << "check: The source chiplet is: " << src_chiplet << std::endl;
            unsigned int dest_pe = p->dst.id; 
            //std::cout << "check: The destination PE is: " << dest_pe << std::endl;
            unsigned int dest_chiplet = (node_to_chiplet(dest_pe)); //This is for single chiplet or input feature transaction only, as weight kernel is transmitted to all chiplets
            //FUNCTION_INFO("Destination Chiplet, Node ID", dest_chiplet, dest_pe );
            //std::cout << "check: The destination chiplet is: " << dest_chiplet << std::endl;
            int data_type = p->dataType;
            //std::cout << "check: The type flag is: " << data_type << std::endl;
            int type_flag;
            std::string flag_type;
            if(data_type == 0){ // '0' is the data transmmission from one PE to another specific PE
                type_flag = 0;
                flag_type = "PE_To_PE";              
            } else if(data_type == 1){ //'1' is input feature which is sent by single chiplet transmission from GB to every PE on a specific chiplet
                type_flag = 1;
                flag_type = "Input_Feature";
            } else if(data_type == 2){ //'2' represents weight kernel, which is sent as cross chiplet transaction from GB to specific PE on every chiplet
                type_flag = 2;
                flag_type = "Weight_Kernel";
            } else if(data_type == 3){ //'3' is OF received as unicast transmission from the PEs to GB and given to NI of GB
                type_flag = 3;
                flag_type = "TYPE_OF";
            }

            int ni_gb_address = 0x400 + (ni_gb_id);  // Address for NI to GB transaction
            trans = new PESendPayload(src_pe, src_chiplet, dest_pe, dest_chiplet, current_flit, sc_time_stamp(), type_flag);
            trans->set_address(ni_gb_address);

            log_info(std::to_string(ni_gb_id) + " (of GB): Processing packet flit with ID: " + std::to_string(current_flit->id) 
                    + " of type: " + flag_type + " for Global Buffer to send to destination: " + std::to_string(p->dst.id)  
                    + " at time: " + trans->timestamp.to_string());
                            
            // Set up the rest of the transaction
            trans->update_timestamp(sc_time_stamp());  // Set the current simulation time
            trans->set_data_length(p->size);
            trans->set_streaming_width(4);
            trans->set_byte_enable_ptr(nullptr);
            trans->set_dmi_allowed(false);
            trans->set_command(TLM_WRITE_COMMAND);
            trans->set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
        } //Initial if - retry condition ends here as packets are generated if no retry required
        tlm::tlm_phase phase = tlm::BEGIN_REQ; // Transaction phase
                
        tlm::tlm_sync_enum status = ni_send_socket->nb_transport_fw(*trans, phase, delay); // Non-blocking transport
        
        if (status == tlm::TLM_COMPLETED || phase == tlm::END_REQ) {
            log_info(std::to_string(ni_gb_id) + " (of GB): Transaction successful, flit sent ID: " 
                    + std::to_string(trans->data->id) + " at time: " + sc_time_stamp().to_string() 
                    + " to Global Buffer");
            //delete trans; // Clean up the transaction
            retry = false;  // Move on to the next transaction
        } else if (status == tlm::TLM_ACCEPTED) {
            log_info(std::to_string(ni_gb_id) + " (of GB): Transaction accepted, waiting for completion.");
            wait(0, SC_NS);  // Wait for response
            retry = false;  // No need to retry, accepted transaction will continue
        } else if (status == tlm::TLM_UPDATED && phase == tlm::BEGIN_REQ) {
            log_info(std::to_string(ni_gb_id) + " (of GB): FIFO full, retrying...");                  
            retry = true;  // Retry the same transaction
            wait(RETRY_DELAY);
        }             
        
    } while(retry); //when retry is true, the transaction is tried again since fifo is full
    
    wait(0, SC_NS);    
}


void NetworkInterfaceGB::thread() {
    while(true){
        //log_info("send_data_process()");
        if (clk.posedge()) {
            if (!packet_send_queue.empty()) {
                //if (credit_counter != 0){
                send_packet_to_gb_from_ni();
                //}
                //else {
                //   log_info("Waiting for Chiplet!");
                //}
            }
            if (!packet_recv_queue.empty()) {
                if (packetPortContainer->portFlowControlIn.read()) {
                    Packet* p = packet_recv_queue.front();
                    packet_recv_queue.pop();
                    packetPortContainer->portValidOut.write(true);
                    packetPortContainer->portDataOut.write(p);
                    // log arrival of packet
                    std::string start_end_pos[2];
                    getDestination(p, start_end_pos);
                    log_info(std::to_string(ni_gb_id) + " (of GB): Packet with id " + to_string(p->id) + 
                               " arrived at PE of Global Buffer ");
                }
            }
        }
        else if (clk.negedge()) {
            packetPortContainer->portValidOut.write(false);
        }
        wait();
    }
}
// Non-blocking TLM function to receive data from GB 
tlm::tlm_sync_enum NetworkInterfaceGB::ni_nb_transport_fw(tlm::tlm_generic_payload& trans, tlm::tlm_phase& phase, sc_time& delay) {
    if (phase == tlm::BEGIN_REQ) {
        PESendPayload* ni_trans = dynamic_cast<PESendPayload*>(&trans);
        if (!ni_trans) {
            wait(OF_DELAY);
            log_error(std::to_string(ni_gb_id) + " (of GB): Invalid transaction type (not PESendPayload).");
            return tlm::TLM_UPDATED;
        }

        PESendPayload* rx_trans5 = new PESendPayload();
        sc_time current_time = sc_time_stamp();
        //manually copy the transaction payload for future use
        sc_time sent_time = ni_trans->timestamp;
        rx_trans5->data = ni_trans->data;
        rx_trans5->src_pe = ni_trans->src_pe;
        rx_trans5->src_chiplet = ni_trans->src_chiplet;
        rx_trans5->dest_pe = ni_trans->dest_pe;
        rx_trans5->dest_chiplet = ni_trans->dest_chiplet;
        rx_trans5->type_flag = ni_trans->type_flag;
        rx_trans5->update_timestamp(current_time);
        rx_trans5->set_address(ni_trans->get_address());
        rx_trans5->set_command(ni_trans->get_command());
        rx_trans5->set_data_ptr(reinterpret_cast<unsigned char*>(&rx_trans5->data));
        rx_trans5->set_data_length(ni_trans->get_data_length());
        rx_trans5->set_streaming_width(ni_trans->get_streaming_width());
        rx_trans5->set_byte_enable_ptr(ni_trans->get_byte_enable_ptr());
        rx_trans5->set_dmi_allowed(ni_trans->is_dmi_allowed());

        // Print which ni is processing the received data
        log_info(std::to_string(ni_gb_id) + " (of GB): of Global Buffer Processing  transaction "  
                + "and received Output Feature (OF) flit ID: " + std::to_string(ni_trans->data->id));
            
        // Send acknowledgment or status back 
        trans.set_response_status(tlm::TLM_OK_RESPONSE);

        // Log the received data and remove the payload
         receive_and_process_flit(*rx_trans5);
        //delete ni_trans;
        return tlm::TLM_COMPLETED;
    }
    return tlm::TLM_ACCEPTED;
    
}


void NetworkInterfaceGB::receive_and_process_flit(tlm::tlm_generic_payload& trans){
    trans.set_response_status(tlm::TLM_OK_RESPONSE);
    PESendPayload* ni_trans1 = dynamic_cast<PESendPayload*>(&trans);
    //unsigned char* data_ptr = trans.get_data_ptr();
    Flit* received_flit = ni_trans1->data;
    Packet* p = received_flit->packet;
    double time = sc_time_stamp().to_double();

    auto position = find(p->inTransmit.begin(), p->inTransmit.end(), received_flit->id);
    if (position!=p->inTransmit.end())
        p->inTransmit.erase(position);
    p->transmitted.push_back(received_flit->id);

    if (received_flit->type==TAIL || received_flit->type==SINGLE){
        stringstream ss;
        std::string flit_type = received_flit->type==SINGLE ? "Single":"Tail";
        ss << ni_gb_id << " (of GB):Receive Flit " << flit_type << *received_flit;
        log_info(ss.str());
    }
    if (!p->toTransmit.empty() || !p->inTransmit.empty()) {
        stringstream ss;
        ss << ni_gb_id << " (of GB): Received Tail Flit, but still missing flits! "
                << *received_flit;
        log_info(ss.str());
    }

    globalReport.issueNoCOutputDataAmount(sc_time_stamp(),globalResources.bitWidth);
    if (p->toTransmit.empty() && p->inTransmit.empty())
        packet_recv_queue.push(p);

}


//print log messages
void NetworkInterfaceGB::log_info(std::string msg) {
    SC_REPORT_INFO(NI_GB_LOG, msg.c_str());
}

void NetworkInterfaceGB::log_error(std::string msg) {
    SC_REPORT_ERROR(NI_GB_LOG, msg.c_str());
}

void NetworkInterfaceGB::log_fatal(std::string msg) {
    SC_REPORT_FATAL(NI_GB_LOG, msg.c_str());
}