/*******************************************************************************
 * Copyright (C) 2024
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
#include "NetworkInterfacePE.h"
#include <iostream>

#include "ratatoskrUtils/traffic/Flit.h"

using namespace std;

NetworkInterfacePE::NetworkInterfacePE(sc_module_name nm, Node& node, TokenManager& token_ref) : 
        NetworkInterface(nm, node, token_ref), token(token_ref), lastReceivedCreditID(-1), credit_counter(0), resp_in_progress(false)
            {
                setup_logger();
                sc_report_handler::set_actions(NI_LOG, SC_INFO, SC_LOG|SC_DISPLAY);
                oec_SingleChiplet_receive_socket.register_b_transport(this, &NetworkInterfacePE::oec_SingleChiplet_b_transport); //register the b_transport fun to each target socket to receive data from oec
                oec_CrossChiplet_receive_socket.register_b_transport(this, &NetworkInterfacePE::oec_CrossChiplet_b_transport);
                

    try {
        this->id = ni_id;
        this->dbid = rep.registerElement("NetworkInterfacePE", this->id);
        this->node = node;
        this->packetPortContainer = new PacketPortContainer(
                ("NI_PACKET_CONTAINER" + std::to_string(this->id)).c_str());
        //copy(max_pos, max_pos + 3, this->max_pos);
    }
    catch (exception& e) {
        log_error("Initialization Error: " + string(e.what()));
    }
    SC_THREAD(thread);
    sensitive << clk.pos() << clk.neg();

    SC_METHOD(receivePacketFromPE);
    sensitive << packetPortContainer->portValidIn.pos();

}

NetworkInterfacePE::~NetworkInterfacePE() {
    delete packetPortContainer;
    
}

void NetworkInterfacePE::initialize(){
    /* NOT IMPLEMENTED */
}

void NetworkInterfacePE::bind(Connection* con, SignalContainer* sigContIn, SignalContainer* sigContOut) {
    if (con==nullptr)
        packetPortContainer->bind(sigContIn, sigContOut);
}

void NetworkInterfacePE::receiveFlitFromRouter() {
    /* NOT IMPLEMENTED */
}

void NetworkInterfacePE::receivePacketFromPE() {
    if (packetPortContainer->portValidIn.posedge()) {
        //LOG(globalReport.verbose_pe_function_calls, "NI" << this->id << "(Node" << node.id << ")\t\t- receive()");
        Packet* packet = packetPortContainer->portDataIn.read();
        generateFlitsForPacket(packet);
        packet_send_queue.push(packet);
    }
}

void NetworkInterfacePE::getDestination(Packet* p, std::string pos[2]){
    pos[0] = to_string(node.pos.x) + "," +
            to_string(node.pos.y) + "," + to_string(node.pos.z);
    pos[1] = to_string(p->dst.pos.x) + "," +
            to_string(p->dst.pos.y) + "," +
            to_string(p->dst.pos.z);
}

void NetworkInterfacePE::generateFlitsForPacket(Packet* p) {
    int flitsPerPacket = p->size;
    for (int i = 0; i<flitsPerPacket; ++i) {
        FlitType flitType;
        if (flitsPerPacket==1)
            flitType = SINGLE;
        else if (i%flitsPerPacket==0)
            flitType = HEAD;
        else if (i%flitsPerPacket==flitsPerPacket-1)
            flitType = TAIL;
        else
            flitType = BODY;
        int seqNum = i%flitsPerPacket;
        Flit* current_flit = new Flit(flitType, seqNum, p, p->dataType, sc_time_stamp().to_double());
        p->toTransmit.push_back(current_flit->id);
        p->flits.push_back(current_flit);
        
        std::string start_end_pos[2];
        getDestination(p, start_end_pos);
        log_info(std::to_string(ni_id) + " (NI: " + std::to_string(ni_pe_id) + " of Chiplet: " 
                + std::to_string(ni_chiplet_id) + "): " + " Flits generated for packet of id " 
                + to_string(p->id) + " to be sent to " + std::to_string(p->dst.id) );
    }
}

void NetworkInterfacePE::send_packet_to_gb() {
    PESendPayload* trans = nullptr;
        
    sc_time delay = sc_time(0, SC_NS);
    bool retry = false;
    do { //Need a retry mechanism despite of the blocking transport here as the NIs are waiting for token to send flits, and if they are not available, it should not just skip the data and move on to next, but insted should retry again
       if (!retry) { //When retry is false new flits are generated from packets
            Packet* p = packet_send_queue.front();
            // if no flits to transmit, generate them
            if (p->toTransmit.empty()){
                generateFlitsForPacket(p);
            }
            // get current flit to send
            flitID_t f_id = p->toTransmit.front();
            auto iter = find_if(p->flits.begin(), p->flits.end(),
                    [&f_id](Flit* f) { return f->id==f_id; });
            Flit* current_flit = *iter;
            current_flit->injectionTime = sc_time_stamp().to_double();

            auto toDelete_pos = find(p->toTransmit.begin(), p->toTransmit.end(), current_flit->id);
            p->toTransmit.erase(toDelete_pos);
            p->inTransmit.push_back(current_flit->id);

            if (p->toTransmit.empty()) {
                packet_send_queue.pop();
            }

            unsigned int src_pe = p->src.id; 
            int src_chiplet = ((src_pe - PE_OFFSET) / PE_PER_CHIPLET); 
            
            unsigned int dest_pe = p->dst.id; 
            int dest_chiplet = ((dest_pe - PE_OFFSET) /PE_PER_CHIPLET);
            //std::cout << "Check NI-PE: at NI_PE: source chiplet id is: " << src_chiplet << " destination chiplet id is: " << dest_chiplet << std::endl; 
            //std::cout << "Check NI-PE: at NI_PE: source PE ID: " <<p->src.id << " destination PE ID: " << p->dst.id << std::endl;
            int chiplet_index = 0x300 + (src_chiplet);  // Address for wavelength λ8 onward

            int data_type1 = p->dataType;
            //std::cout << "check: The type flag is: " << data_type1 << std::endl;
            int type_flag1;
            std::string flag_type1;
            if(data_type1 == 0){ // '0' is the data transmmission from one PE to another specific PE
                type_flag1 = 0;
                flag_type1 = "PE_To_PE";              
            } else if(data_type1 == 1){ //'1' is input feature which is sent by single chiplet transmission from GB to every PE on a specific chiplet
                type_flag1 = 1;
                flag_type1 = "Input_Feature";
            } else if(data_type1 == 2){ //'2' represents weight kernel, which is sent as cross chiplet transaction from GB to specific PE on every chiplet
                type_flag1 = 2;
                flag_type1 = "Weight_Kernel";
            } else if(data_type1 == 3){ //'3' is OF received as unicast transmission from the PEs to GB and given to NI of GB
                type_flag1 = 3;
                flag_type1 = "TYPE_OF";
            }
            
            trans = new PESendPayload(src_pe, src_chiplet, dest_pe, dest_chiplet, current_flit, sc_time_stamp(), data_type1);
            trans->set_address(chiplet_index);
            log_info(std::to_string(src_pe)+ " (NI: " + std::to_string(ni_pe_id) + " of Chiplet: " 
                    + std::to_string(src_chiplet) + "): "  + " Processing packet flit with ID: " 
                    + std::to_string(trans->data->id) + " and data type: " + flag_type1 + " for Chiplet " 
                    + std::to_string(dest_chiplet) + " and Network Interface " + std::to_string(dest_pe) 
                    + " at time: " + trans->timestamp.to_string());
                            
            // Set up the rest of the transaction
            trans->update_timestamp(sc_time_stamp());  // Set the current simulation time
            trans->set_data_length(p->size);
            trans->set_streaming_width(4);
            trans->set_byte_enable_ptr(nullptr);
            trans->set_dmi_allowed(false);
            trans->set_command(TLM_WRITE_COMMAND);
        } //Initial if - retry condition ends here as packets are generated if no retry required
        if (token.request_token(ni_id, ni_chiplet_id))  {
            log_info(std::to_string(ni_id)+ " (NI: " + std::to_string(ni_pe_id) + " of Chiplet: " 
                    + std::to_string(ni_chiplet_id) + "): " + " Transmitting flits with ID: " 
                    + std::to_string(trans->data->id));
            oec_send_socket ->b_transport(*trans, delay); 
            if(trans->get_response_status() == TLM_OK_RESPONSE) {
                log_info( std::to_string(ni_id) + " (NI: " + std::to_string(ni_pe_id) + " of Chiplet: " 
                        + std::to_string(ni_chiplet_id) + "): " + " Transaction of flit: " 
                        + std::to_string(trans->data->id) + " completed successfully from source NI: " 
                        + std::to_string(ni_id) + " to destination NI ");
            //receive_and_process_flit(*trans);
            } else {
                log_error(std::to_string(ni_id) + " (NI: " + std::to_string(ni_pe_id) + " of Chiplet: " 
                        + std::to_string(ni_chiplet_id) + "): " +  " Error in transmitting flits: " 
                        + std::to_string(trans->data->id) + " from NI to OEC ");
            }
            token.release_token(ni_id, ni_chiplet_id);
            retry = false;
            //delete trans; 
        } else {
            log_info(std::to_string(ni_id)+ " (NI: " + std::to_string(ni_pe_id) + " of Chiplet: " 
                    + std::to_string(ni_chiplet_id) + "): " + ": Token of the Chiplet " 
                    + std::to_string(ni_chiplet_id) + " is unavailable. Retrying with data: " 
                    + std::to_string(trans->data->id));
            retry = true;
            wait(RETRY_DELAY);  // Wait before retrying
            //std::cout <<"Retrying data transmission for network interface " << ni_id << " of chiplet id " << ni_chiplet_id << std::endl;
        }              
        
    } while(retry); //when retry is true, the transaction is tried again since fifo is full

    wait(0, SC_NS);    
}


void NetworkInterfacePE::thread() {
    while(true){
        //log_info("send_data_process()");
        if (clk.posedge()) {
            if (!packet_send_queue.empty()) {
                //if (credit_counter != 0){
                send_packet_to_gb();
                //}
                //else {
                //   log_info("Waiting for Chiplet!");
                //}
            }
            if (!packet_recv_queue.empty()) {
                if (packetPortContainer->portFlowControlIn.read()) {
                    Packet* p = packet_recv_queue.front();
                    packet_recv_queue.pop();
                    packetPortContainer->portValidOut.write(true);
                    packetPortContainer->portDataOut.write(p);
                    // log arrival of packet
                    std::string start_end_pos[2];
                    getDestination(p, start_end_pos);
                    log_info(std::to_string(ni_id)+ " (NI: " + std::to_string(ni_pe_id) + " of Chiplet: " 
                            + std::to_string(ni_chiplet_id) + "): " + " Packet with id " 
                            + std::to_string(p->id) + " arrived at Parent PE ");//start_end_pos[1]);
                }
            }
        }
        else if (clk.negedge()) {
            packetPortContainer->portValidOut.write(false);
        }
        wait();
    }
}

//To receive single chiplet transactions from the GB
void NetworkInterfacePE::oec_SingleChiplet_b_transport(tlm_generic_payload& trans, sc_time& delay) {
    PESendPayload* orig_trans = dynamic_cast<PESendPayload*>(&trans);
    if (!orig_trans) {
        wait(IF_DELAY);
        log_error(std::to_string(ni_id) + " Invalid transaction type (not PESendPayload).");
        trans.set_response_status(TLM_GENERIC_ERROR_RESPONSE);
        return;
    }

    PESendPayload* pe_trans1 = new PESendPayload();

    sc_time current_time = sc_time_stamp();
    //manually copy the transaction payload for future use
    sc_time sent_time = orig_trans->timestamp;
    pe_trans1->data = orig_trans->data;
    pe_trans1->src_pe = orig_trans->src_pe;
    pe_trans1->src_chiplet = orig_trans->src_chiplet;
    pe_trans1->dest_pe = orig_trans->dest_pe;
    pe_trans1->dest_chiplet = orig_trans->dest_chiplet;
    pe_trans1->type_flag = orig_trans->type_flag;
    pe_trans1->update_timestamp(current_time);
    pe_trans1->set_address(orig_trans->get_address());
    pe_trans1->set_command(orig_trans->get_command());
    pe_trans1->set_data_ptr(reinterpret_cast<unsigned char*>(&pe_trans1->data));
    pe_trans1->set_data_length(orig_trans->get_data_length());
    pe_trans1->set_streaming_width(orig_trans->get_streaming_width());
    pe_trans1->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
    pe_trans1->set_dmi_allowed(orig_trans->is_dmi_allowed());

    int address = pe_trans1->get_address(); // address is used to determine wavelength
    trans.set_response_status(TLM_OK_RESPONSE);  // Set response status
    if ((address >= 0x100 && address < 0x200)) {
        // Single-chiplet communication
        int chiplet_index = address - 0x100;
        int chiplet_id = chiplet_index - PE_PER_CHIPLET;
        
        int single_wavelength = chiplet_index;
        //log_info(std::to_string(ni_id) + " Single-chiplet communication, wavelength: " + std::to_string(single_wavelength));
        wait(IF_DELAY);
        // Single-chiplet communication (λM)
        log_info(std::to_string(ni_id)+ " (NI: " + std::to_string(ni_pe_id) + " of Chiplet: " 
                + std::to_string(ni_chiplet_id) + "): " + " Received Single Chiplet transaction: " 
                + std::to_string(pe_trans1->data->id) + " with Wavelength: " + std::to_string(single_wavelength));
    } else {

        log_error(std::to_string(ni_id) + " Invalid or unrecognized destination: " + std::to_string(pe_trans1->dest_pe));
    
    }
    
    //SINGLE_FLIT_COUNTER++;
    receive_and_process_flit(*pe_trans1);
    //delete pe_trans1;  // Clean up dynamically allocated memory
    
}


// Handle incoming Cross chiplet transaction from the Global Buffer
void NetworkInterfacePE::oec_CrossChiplet_b_transport(tlm_generic_payload& trans, sc_time& delay) {
    PESendPayload* orig_trans = dynamic_cast<PESendPayload*>(&trans);
    if (!orig_trans) {
        wait(WK_DELAY);
        log_error(std::to_string(ni_id) + " Invalid transaction type (not PESendPayload).");
        trans.set_response_status(TLM_GENERIC_ERROR_RESPONSE);
        return;
    }

  PESendPayload* pe_trans2 = new PESendPayload();
  sc_time current_time = sc_time_stamp();
    //manually copy the transaction payload for future use
    sc_time sent_time = orig_trans->timestamp;
    pe_trans2->data = orig_trans->data;
    pe_trans2->src_pe = orig_trans->src_pe;
    pe_trans2->src_chiplet = orig_trans->src_chiplet;
    pe_trans2->dest_pe = orig_trans->dest_pe;
    pe_trans2->dest_chiplet = orig_trans->dest_chiplet;
    pe_trans2->type_flag = orig_trans-> type_flag;
    pe_trans2->update_timestamp(current_time);
    pe_trans2->set_address(orig_trans->get_address());
    pe_trans2->set_command(orig_trans->get_command());
    pe_trans2->set_data_ptr(reinterpret_cast<unsigned char*>(&pe_trans2->data));
    pe_trans2->set_data_length(orig_trans->get_data_length());
    pe_trans2->set_streaming_width(orig_trans->get_streaming_width());
    pe_trans2->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
    pe_trans2->set_dmi_allowed(orig_trans->is_dmi_allowed());
    // Complete the original transaction
    trans.set_response_status(TLM_OK_RESPONSE);
    int address = pe_trans2->get_address(); // address is used to determine wavelength
    if ((address >= 0x200) && ((pe_trans2->dest_pe == ni_id)||pe_trans2->type_flag == 2 )) {
        int cross_wavelength = address - 0x200;
        //log_info("Cross-chiplet communication, wavelength : " + std::to_string(cross_wavelength));
        wait(WK_DELAY);
        // Cross-chiplet communication (λ0, λ1, λ2, ...)
        log_info(std::to_string(ni_id)+ " (NI: " + std::to_string(ni_pe_id) + " of Chiplet: " 
                + std::to_string(ni_chiplet_id) + "): " + " Received cross-chiplet data: " 
                + std::to_string(pe_trans2->data->id) + " with Wavelength: " + std::to_string(cross_wavelength));
    } else {

        log_info(std::to_string(ni_id)+ " (NI: " + std::to_string(ni_pe_id) + " of Chiplet: " 
                + std::to_string(ni_chiplet_id) + "): " + " Invalid or unrecognized destination: " 
                + std::to_string(pe_trans2->dest_pe) + " still forwarding the packet to parent PE");
            
    }
    //log_info(std::to_string(ni_id) + " Calling function to process flit for data " + std::to_string(pe_trans2->data->id));
    receive_and_process_flit(*pe_trans2);
    //delete pe_trans2;  // Clean up dynamically allocated memory
    
}



void NetworkInterfacePE::receive_and_process_flit(tlm::tlm_generic_payload& trans){
    trans.set_response_status(tlm::TLM_OK_RESPONSE);
    PESendPayload* orig_trans = dynamic_cast<PESendPayload*>(&trans);
    //unsigned char* data_ptr = trans.get_data_ptr();
    
    Flit* received_flit = orig_trans->data;//reinterpret_cast<Flit*>(data_ptr);
    
    Packet* p = received_flit->packet;
    double time = sc_time_stamp().to_double();
    
    received_flit->flit_counter++;
    //std::cout << "Check NI: The id of the flit: " << received_flit->id << " & The counter value: " << received_flit->flit_counter << std::endl;
    auto position = find(p->inTransmit.begin(), p->inTransmit.end(), received_flit->id);
    if (position != p->inTransmit.end()) {
        if (received_flit->flit_counter == NO_OF_TRANSACTION) {
            p->inTransmit.erase(position);
            log_info(std::to_string(ni_id)+ " (NI: " + std::to_string(ni_pe_id) + " of Chiplet: " 
                    + std::to_string(ni_chiplet_id) + "): " + " Erassed the received flit: " 
                    + std::to_string(received_flit->id) + " from the inTransmit vector" );
        }
    }
    p->transmitted.push_back(received_flit->id);

    if (received_flit->type==TAIL || received_flit->type==SINGLE){
        stringstream ss;
        std::string flit_type = received_flit->type==SINGLE ? "Single":"Tail";
        ss << ni_id << " (NI: " << ni_pe_id << " of Chiplet: " 
                << ni_chiplet_id << "): " << " Receive Flit " << flit_type << *received_flit;
        log_info(ss.str());
    }
    if (!p->toTransmit.empty() || !p->inTransmit.empty()) {
        stringstream ss;
        ss << ni_id << " (NI: " << ni_pe_id << " of Chiplet: " << ni_chiplet_id << "): " 
        << " Received Tail Flit, but still missing flits! "<< *received_flit;
        log_info(ss.str());
    }

    
    // Increment the packet's global counter
    packet_recv_queue.push(p);
        log_info(std::to_string(ni_id)+ " (NI: " + std::to_string(ni_pe_id) + " of Chiplet: " 
                + std::to_string(ni_chiplet_id) + "): " + " Pushed packet: " + std::to_string(p->id) 
                + " into packet receive queue" );
    // std::cout << "Check NI: The id of Packet: " << p->id << " & The counter value: " 
    //         << p->packet_counter << std::endl;
    globalReport.issueNoCOutputDataAmount(sc_time_stamp(),globalResources.bitWidth);
}


//print log messages
void NetworkInterfacePE::log_info(std::string msg){
    SC_REPORT_INFO(NI_LOG, msg.c_str());
}

void NetworkInterfacePE::log_error(std::string msg){
    SC_REPORT_ERROR(NI_LOG, msg.c_str());
}

void NetworkInterfacePE::log_fatal(std::string msg){
    SC_REPORT_FATAL(NI_LOG, msg.c_str());
}

