/*******************************************************************************
 * Copyright (C) 2024
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/

#include "NetworkManager.h"
#include "utils/noc_logger.h"

NetworkManager::NetworkManager(sc_module_name nm, std::string configFolder) : sc_module(nm) {
    setup_logger();
    sc_report_handler::set_actions(NM_LOG, SC_INFO, SC_LOG|SC_DISPLAY);
    dbid = rep.registerElement("NetworkManager", 0);
    //tlmNoc = new NocTlm("NocTlm", configFolder);
    tlmNoc = std::make_unique<NocTlm>("NocTlm", configFolder);
    //networkParticipants.resize(globalResources.nodes.size());
    createTrafficPool();
    createClocks();
    createNetworkParticipants();
    //createLinks();
    runNoC();
}
// NetworkManager::~NetworkManager() {
// }
void NetworkManager::createClocks() {
    
    clocks.resize(globalResources.nodeTypes.size());
    for (const auto &nodeType : globalResources.nodeTypes) {
        clocks.at(nodeType->id) = std::make_unique<sc_clock>(
                ("NodeType" + std::to_string(nodeType->id) + "Clock").c_str(),
                nodeType->clockDelay, SC_NS);
            //std::cout << "node type " << nodeType->id << std::endl; 
        log_info("Created clocks for id: " + std::to_string(nodeType->id));
    }
}


void NetworkManager::createTrafficPool() {
    unsigned long numOfPEs =  globalResources.nodes.size();

    log_info("The total number of PEs: " + std::to_string(numOfPEs));
    //std::cout << "Derived numOfPEs: " << numOfPEs << std::endl;

    #ifndef ENABLE_NETRACE
        if (globalResources.benchmark == "task") {
            
            tp = std::make_unique<TaskPool>();
            //std::cout << "This is a task: " << std::endl;   
        } else if (globalResources.benchmark == "synthetic") {
            
            tp = std::make_unique<SyntheticPool>();
        } else {
            FATAL("Please specify correct benchmark type");
        }
    #endif
    // #ifdef ENABLE_NETRACE
    //     tp = std::make_unique<NetracePool>("NetracePool");
    // #endif
    tp->processingElements.resize(numOfPEs);    
}

void NetworkManager::createNetworkParticipants() {
    
    int numOfPEs = tp->processingElements.size();
    //std::cout << "The number of Nodess from processingElement class size " << numOfPEs << std::endl;

    for (Node &n : globalResources.nodes) {
        if (n.type->model == "ProcessingElement") {
            int id = (n.id - PE_OFFSET) / PE_PER_CHIPLET;
            //std::cout << "The id from n.id equation is: " << id << std::endl;
            int j = (n.id - PE_OFFSET) % PE_PER_CHIPLET;
            //std::cout << "The j from n.id equation is: " << j << std::endl;
            int index = tlmNoc->nis[id][j]->ni_id;
            //std::cout << "The  value of index derived from the number of PEs from processingElement class size " << index << std::endl;
            NetworkInterfacePE* ni = tlmNoc->nis[id][j];
            if (!ni) continue; // Ensure ni is valid

            std::string pe_name = "pe_" + std::to_string(n.id) + "_" + std::to_string(j);
            //ProcessingElementVC *pe = new ProcessingElementVC(pe_name.c_str(), n, tp.get(), index);
            auto pe = std::make_unique<ProcessingElementVC>(pe_name.c_str(), n, tp.get(), index);
            auto sig1 = std::make_unique<PacketSignalContainer>(
                ("packetSigCon1_" + std::to_string(n.id) + "_" + std::to_string(j)).c_str());
            auto sig2 = std::make_unique<PacketSignalContainer>(
                ("packetSigCon2_" + std::to_string(n.id) + "_" + std::to_string(j)).c_str());

            ni->bind(nullptr, sig1.get(), sig2.get());
            pe->bind(nullptr, sig2.get(), sig1.get());
            ni->clk(*clocks.at(n.type->id));

            if (index < tp->processingElements.size()) {
                tp->processingElements[index] = pe.get();
            }

            // networkParticipants.push_back(dynamic_cast<NetworkParticipant *>(pe));
            // packetSignalContainers.push_back(move(sig1));
            // packetSignalContainers.push_back(move(sig2));
            networkParticipants.push_back(std::move(pe));
            packetSignalContainers.push_back(std::move(sig1));
            packetSignalContainers.push_back(std::move(sig2));
   
        } else if (n.type->model == "GlobalProcessingElement") {
            int id = n.id; 
            int index = tlmNoc->ni_gb->ni_gb_id;
            NetworkInterfaceGB* ni_gb = tlmNoc->ni_gb;
            if (!ni_gb) continue; // Ensure ni is valid
            std::string pe_gb_name = "pe_gb_" + std::to_string(n.id);
            auto pe_gb = std::make_unique<ProcessingElementVC>(pe_gb_name.c_str(), n, tp.get(), index);
           // std::cout << "Created the processing element for Global Buffer " << pe_gb_name << std::endl;
            
            auto sig1 = std::make_unique<PacketSignalContainer>(
                ("packetSigCon1_gb_" + std::to_string(n.id) + "_").c_str());
            auto sig2 = std::make_unique<PacketSignalContainer>(
                ("packetSigCon2_gb_" + std::to_string(n.id) + "_").c_str());

            ni_gb->bind(nullptr, sig1.get(), sig2.get());
            pe_gb->bind(nullptr, sig2.get(), sig1.get());
            ni_gb->clk(*clocks.at(n.type->id));

            //std::cout<<"tp pe size: "<<tp->processingElements.size()<<std::endl;
            //std::cout<<"index: "<<index<<std::endl;
            if (index < tp->processingElements.size()) {
                tp->processingElements[index] = pe_gb.get();
            }

            networkParticipants.push_back(std::move(pe_gb));
            packetSignalContainers.push_back(std::move(sig1));
            packetSignalContainers.push_back(std::move(sig2));
        }
    }
}



void NetworkManager::runNoC() {
    //std::cout << "Starting at the Network Manager " << std::endl;
    for (auto &r : networkParticipants) {
        if (r) {
            r->initialize();
        }
    }
    tp->start();
}

 NetworkManager::~NetworkManager() {
    for (auto &r : networkParticipants)
        r.reset();
    for (auto &c : clocks)
        c.reset();
    networkParticipants.clear();  
    clocks.clear();  
    packetSignalContainers.clear();
    tp.reset();
    tlmNoc.reset();  
}


// NetworkManager::~NetworkManager() {
//     for (auto &r : networkParticipants)
//         delete r;
//     networkParticipants.clear();
// }

void NetworkManager::log_info(std::string msg){
    SC_REPORT_INFO(NM_LOG, (msg).c_str());
}