/*******************************************************************************
 * Copyright (C) 2024
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/

// NocTlm.cpp
#include "noc.h"

NocTlm::NocTlm(sc_module_name nm, std::string config_folder): sc_module(nm) {
    setup_logger();
    sc_report_handler::set_actions(NOC_Set_LOG, SC_INFO, SC_LOG|SC_DISPLAY);
    initializeGlobalResources(config_folder);
    setup_modules();
    
}
// NocTlm::~NocTlm() {
//     std::exit(0);
// }
void NocTlm::initializeGlobalResources(std::string config_folder) {
    // initialize global resources and rep
    globalResources = GlobalResources::getInstance();
    GlobalReport &globalReport = GlobalReport::getInstance();
    Report &rep = Report::getInstance();
    sleep(1);

    std::string config_path = "config/"+config_folder+"/config.xml"; //"config/config.xml";
    globalResources.readConfigFile(config_path);
    globalReport.readConfigFile(config_path);

    std::string network_path = "config/"+config_folder+"/net.xml";
    //globalResources.noc_file;
    globalResources.readNoCLayout(network_path);
    globalResources.readTaskAndMapFiles(globalResources.data_file, globalResources.map_file);
    //globalReport.resizeMatrices();

    globalResources.activateFlitTracing = false;
    globalResources.outputDirectory = "out";
}

void NocTlm::setup_modules() {
    for (Node &n : globalResources.nodes) {
        if (n.type->model == "GlobalBuffer"){
            int gb_id = n.id;
            log_info("The id of the Global Buffer is: " + std::to_string(gb_id));
            //std::cout << "The id of the Global Buffer is: " << gb_id << std::endl;
            // Allocate memory for GlobalBuffer
            gb = new GlobalBuffer("GlobalBuffer", gb_id, CHIPLET_NO, PE_PER_CHIPLET);
        }
    }

    for (Node &n : globalResources.nodes) {
        if (n.type->model == "Chiplet") {
            int id = n.id;
            log_info("Created the chiplet with internal id: " + std::to_string(id));
            //std::cout << "The node ID of chiplet is: " << n.id << std::endl;
            std::string chip_name = "ChipletInterface_" + std::to_string(id);
            chiplets[id] = new ChipletInterface(chip_name.c_str(), PE_PER_CHIPLET);
            chiplets[id] -> chiplet_id = id;
            //std::cout << "Created the chiplet with id: " << id << std::endl;
        }
    }

    for (Node &n : globalResources.nodes) {
        if (n.type->model == "Chiplet"){
            int id = n.id;
            std::string name = "InterposerInterface_" + std::to_string(id);
            interposers[id] = new InterposerInterface(name.c_str());
            interposers[id] -> interposer_id = id;
            log_info("Created the interposer interface with internal id: " + std::to_string(id));
            gb->SingleChiplet_from_GB_sockets[id].bind(interposers[id]->gb_SingleChiplet_receive_socket);
            gb->CrossChiplet_from_GB_sockets[id].bind(interposers[id]->gb_CrossChiplet_receive_socket);
            interposers[id]->gb_send_socket.bind(gb->receive_to_GB_sockets[id]);
            log_info("Bound interposer interface sockets to global buffer sockets");
            // Bind chiplet sockets 
            interposers[id]->chiplet_SingleChiplet_send_socket.bind(chiplets[id]->interposer_SingleChiplet_receive_socket);
            interposers[id]->chiplet_CrossChiplet_send_socket.bind(chiplets[id]->interposer_CrossChiplet_receive_socket);
            chiplets[id]->interposer_send_socket.bind(interposers[id]->chiplet_receive_socket);
            log_info("Bound interposer interface sockets to Chiplet interface sockets");

            std::string token_name = "TokenManager_" + std::to_string(id);
            tokens[id] = new TokenManager(token_name.c_str());  // Create a token for each chiplet
            if (tokens[id] == nullptr) {
                log_error("Failed to create TokenManager for chiplet " + std::to_string(id));
            }
            log_info("Created the tokens for the PEs");

            for (int j = 0; j < PE_PER_CHIPLET; ++j) {
                std::string ni_name = "ni" + std::to_string(id) + "_" + std::to_string(j);
                nis[id][j] = new NetworkInterfacePE(ni_name.c_str(), n, *tokens[id]);
                nis[id][j]->ni_chiplet_id = id;  // Set source router ID
                nis[id][j]->ni_id = PE_PER_CHIPLET*id + j + PE_OFFSET;  // Set source NI ID
                nis[id][j]->ni_pe_id = j;
                //std::cout << "ni is created " << nis[id][j]->ni_id << std::endl;
                log_info("Created the network interface with internal id: " + std::to_string(nis[id][j]->ni_id));
                std::string oe_name = "O_E_converter_" + std::to_string(id) + "_" + std::to_string(j);
                oec[id][j] = new O_E_converter(oe_name.c_str());
                oec[id][j] -> O_E_converter_id = j;
                log_info("Created the optical to electrical converters with internal id: " + std::to_string(j));
                chiplets[id]->oec_SingleChiplet_send_sockets[j].bind(oec[id][j]->ci_SingleChiplet_receive_socket);
                chiplets[id]->oec_CrossChiplet_send_sockets[j].bind(oec[id][j]->ci_CrossChiplet_receive_socket);
                oec[id][j]->ci_send_socket.bind(chiplets[id]->oec_receive_sockets[j]);
                log_info("Bound Optical to Electrical converter sockets to Chiplet Interface sockets");
                
                oec[id][j]->ni_SingleChiplet_send_socket.bind(nis[id][j] ->oec_SingleChiplet_receive_socket);
                oec[id][j]->ni_CrossChiplet_send_socket.bind(nis[id][j] ->oec_CrossChiplet_receive_socket);
                nis[id][j] ->oec_send_socket.bind(oec[id][j] ->ni_to_oec_receive_socket);
                log_info("Bound Optical to Electrical converter sockets to Network Interface sockets");                
                //std::cout << "Created modules and binded sockets" << std::endl;
            }   
        }
    }

    for (Node &n : globalResources.nodes) {
        if (n.type->model == "GlobalProcessingElement"){
            int ni_id = n.id;
            ni_gb = new NetworkInterfaceGB("NetworkInterfaceGB", n, *tokens[0]); //The token for Network interface of GB is unnecessary, but since we passed the token reference in the ratatoskr NI, we are simply adding the reference here
            ni_gb -> ni_gb_id = n.id;
            gb->gb_to_ni_send_socket.bind(ni_gb->ni_receive_socket);
            ni_gb->ni_send_socket.bind(gb->ni_to_gb_receive_socket);
        }
    }
}

NocTlm::~NocTlm() {
    //Clean up memory to avoid leaks
    if (gb != nullptr) {
        delete gb; // Free the memory
        delete ni_gb;
        gb = nullptr;
        ni_gb = nullptr;
    }
    for (int i = 0; i < CHIPLET_NO; ++i) {
        delete interposers[i];
        delete chiplets[i];
        delete tokens[i];
        for (int j = 0; j < PE_PER_CHIPLET; ++j) {
            delete oec[i][j];
            delete nis[i][j];
        }   
    
    }
}

void NocTlm::log_info(std::string msg){
    SC_REPORT_INFO(NOC_Set_LOG, (msg).c_str());
}

void NocTlm::log_warn(std::string msg){
    SC_REPORT_WARNING(NOC_Set_LOG, (msg).c_str());
}

void NocTlm::log_error(std::string msg){
    SC_REPORT_ERROR(NOC_Set_LOG, (msg).c_str());
}