/*******************************************************************************
 * Copyright (C) 2024
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/

#pragma once

#include <systemc.h>
#include <tlm.h>
#include <tlm_utils/simple_initiator_socket.h>
#include <tlm_utils/simple_target_socket.h>
#include "ratatoskrUtils/utils/Structures.h"
#include <ratatoskrUtils/utils/TrafficTracer.h>
#include <ratatoskrUtils/utils/GlobalResources.h>

using namespace sc_core;  // For sc_time

// Global parameters to generalize the design
#define PE_NO 65 // Total Number of PEs
#define CHIPLET_NO 8 
#define PE_PER_CHIPLET 8 //(PE_NO/CHIPLET_NO)  Number of NIs per chiplet
#define INTERPOSER_INTERFACE 8 // 
#define NUM_MATRICES 10
#define PE_OFFSET (CHIPLET_NO)
#define LOG_NAME "MY_LOG"
#define TYPE_PE_TO_PE 0
#define TYPE_IF 1
#define TYPE_WK 2
#define TYPE_OF 3
#define SINGLE_CHIPLET_FIFO_CAPACITY 1000 //capacity of FIFO in GB
#define CROSS_CHIPLET_FIFO_CAPACITY 1000 //capacity of FIFO in GB
#define OF_FIFO_CAPACITY 1000 //capacity of FIFO in GB
#define NO_OF_TRANSACTION 8
#define WK_DELAY sc_time(0.5, SC_NS) //The delay for processing weight kernels as cross chiplet transactions
#define RETRY_DELAY sc_time(2, SC_NS) // The delay for retry (tokens, FiFo full case..)
#define OF_DELAY sc_time(0.5, SC_NS) //Delay for transmitting ouput features from PE to GB
#define UPDATE_DELAY sc_time(1, SC_NS) // Delay in returning statuses in TLM non-blocking transport
#define IF_DELAY sc_time(0.5, SC_NS) //Delay in transmitting input features in single-chiplet transactions


// Inline function to convert node_id to PE_id
inline int node_to_pe(int node_id) {
    return (node_id - PE_OFFSET) % PE_PER_CHIPLET;
}

// Inline function to find chiplet id from PE node id
inline int node_to_chiplet(int node_id) {
    return (node_id - PE_OFFSET) / PE_PER_CHIPLET;
}
//Usage: FUNCTION_INFO("x, y, z", x, y, z):- This function is to print the debug messages for the program
#define FUNCTION_INFO(var_names, ...) { \
    std::cout << "------------------------------Debug start---------------------------------------" << std::endl; \
    std::cout << "|Function: " << __func__ << ", Line: " << __LINE__ << ", Class: " << typeid(*this).name() << std::endl; \
    std::cout << "|Variables: " << var_names  << std::endl; \
    printVariables(__VA_ARGS__); \
}

  // Function to print the variables and their values
  template<typename... Args>
  void printVariables(Args... args) {
      std::string values = "[";
      ((values += std::to_string(args) + ", "), ...);  // Unpacking the args into a string

      // Remove the last comma and space if present
      if (values.size() > 1) values = values.substr(0, values.size() - 2);
      values += "]";

      std::cout << "|Values: " << values << std::endl;
      std::cout << "-----------------------------------Debug end------------------------------------" << std::endl;
  }


struct CustomPayload : public tlm::tlm_generic_payload {
    
  unsigned int num_rows;           // Number of rows in the matrix
  unsigned int num_cols;           // Number of columns in the matrix
  int data;           // Flattened matrix data
  sc_core::sc_time timestamp;      // Timestamp field

  // Default constructor
  CustomPayload()
      : num_rows(0),
        num_cols(0),
        data(0),
        timestamp(sc_core::SC_ZERO_TIME) {
      set_data_ptr(reinterpret_cast<unsigned char*>(&data));
      set_data_length(sizeof(data));
      set_streaming_width(sizeof(data));
  }

  // Constructor with random data
  CustomPayload(unsigned int num_rows, unsigned int num_cols, int data, sc_core::sc_time time = sc_core::SC_ZERO_TIME)
    : num_rows(num_rows),
      num_cols(num_cols),
      data(data),
      timestamp(time) {
    
      set_data_ptr(reinterpret_cast<unsigned char*>(&data));
      set_data_length(sizeof(data));
      set_streaming_width(sizeof(data));    
  }

  // Function to update the timestamp at any stage
  void update_timestamp(sc_core::sc_time new_time) {
      timestamp = new_time;
  }
};

struct PESendPayload : public tlm::tlm_generic_payload {

  unsigned int src_chiplet;
  unsigned int src_pe;
  unsigned int dest_chiplet;           // Number of rows in the matrix
  unsigned int dest_pe;           // Number of columns in the matrix
  Flit* data;           // Flattened matrix data
  sc_core::sc_time timestamp;      // Timestamp field
  int type_flag;

  // Default constructor
  PESendPayload()
    : src_pe(0),
      src_chiplet(0),
      dest_pe(0),
      dest_chiplet(0),
      data(nullptr),
      timestamp(sc_core::SC_ZERO_TIME),
      type_flag(0) {
    set_data_ptr(reinterpret_cast<unsigned char*>(&data));
    set_data_length(sizeof(data));
    set_streaming_width(sizeof(data));
  }

  // Constructor with random data
  PESendPayload(unsigned int src_pe, unsigned int src_chiplet, unsigned int dest_pe, unsigned int dest_chiplet, 
                Flit* flit_data, sc_core::sc_time time = sc_core::SC_ZERO_TIME, int type_flag = 0) 
    : src_pe(src_pe),
      src_chiplet(src_chiplet),
      dest_pe(dest_pe),
      dest_chiplet(dest_chiplet),
      data(flit_data),
      timestamp(time),
      type_flag (type_flag) {
      set_data_ptr(reinterpret_cast<unsigned char*>(&data));
      set_data_length(sizeof(data));
      set_streaming_width(sizeof(data));
      
  }

  // Function to update the timestamp at any stage
  void update_timestamp(sc_core::sc_time new_time) {
      timestamp = new_time;
  }
};  
