/*******************************************************************************
 * Copyright (C) 2024
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
#ifndef TOKEN_MANAGER_H
#define TOKEN_MANAGER_H

#include <systemc>
#include <vector>
#include <queue>
#include <iostream>
#include <unordered_set>
#include "configuration.h"

using namespace sc_core;

class TokenManager : public sc_module {
public:
    TokenManager(sc_module_name name) : sc_module(name) {
        token_owner = -1;      // Start with all tokens unowned
        token_status = false; // All tokens are initially free
        last_token_owner = -1; // Initialize last owner to -1
        waiting_queue.reserve(PE_PER_CHIPLET); // Reserve space for waiting_queue
        waiting_set = {}; // Initialize an empty unordered_set
        sc_report_handler::set_log_file_name("out/report.log");
        sc_report_handler::set_actions(LOG_NAME, SC_INFO, SC_LOG|SC_DISPLAY);
    }

    // Request token function
    bool request_token(int ni_id, int token_id) {
        if (token_owner == ni_id || (token_owner == -1 && waiting_queue.empty())) {
            //std::cout << "current owner of token before assigning is " << token_owner << std::endl;
           
            log_info("Token for the chiplet " + std::to_string(token_id) + " granted to PE " + std::to_string(ni_id));
            token_owner = ni_id;
            //std::cout << "current owner of token after assigning is " << token_owner << std::endl;
            token_status = true;  // Mark token as in use
            //std::cout << "current status of token after assigning is " << token_status << std::endl;
            last_token_owner = ni_id;  // Update last owner
            return true;
        } else {
            // Token is occupied, check waiting list
            if (waiting_set.find(ni_id) == waiting_set.end()) {
                
                // NI is not already in the waiting queue
                waiting_queue.push_back(ni_id);
                waiting_set.insert(ni_id); // Add to the set for quick lookup
                log_info("NI " + std::to_string(ni_id) + " added to queue for token " + std::to_string(token_id));
            }
            return false; // Token not granted
        }
    }

    bool release_token(int ni_id, int token_id) {
        if (token_owner != ni_id) {
            log_error("NI " + std::to_string(ni_id) + " attempted to release a token it does not own.");
            return false; // Indicate the release failed
        }

        log_info("NI " + std::to_string(ni_id) + " released token for chiplet " + std::to_string(token_id));
        if (!waiting_queue.empty()) {
            int next_ni = -1;

            int local_ni = (ni_id - PE_OFFSET) % PE_PER_CHIPLET; // Local ID of current NI within the chiplet
            int ref_start_ni = ni_id - local_ni;
            //std::cout << "The ID of the first NI in the current chiplet " << ref_start_ni << std::endl;
            //std::cout << "The local NI value calculated for round-robin release of token is " << local_ni << std::endl;
            int start_ni = (local_ni + 1) % PE_PER_CHIPLET;      // Start clockwise from the next PE
            //std::cout << "The start NI value calculated for round-robin release of token is " << start_ni << std::endl;

            // Find the next NI in the waiting queue in clockwise order
            for (int i = 0; i < PE_PER_CHIPLET; ++i) {
                int ref_ni = (start_ni + i) % PE_PER_CHIPLET; // Global ID of candidate NI
                //std::cout << "The reference NI value calculated by round-robin method to assign the token is next " << ref_ni << std::endl;
                int candidate_ni = ref_start_ni + ref_ni;
                //std::cout << "The candidate NI value calculated for round-robin release of token is " << candidate_ni << std::endl;
                if (waiting_set.find(candidate_ni) != waiting_set.end()) {
                    next_ni = candidate_ni;
                    //std::cout << "The next NI value to which the token is given in round-robin  is " << next_ni << std::endl;
                    break;
                }
            }

            if (next_ni != -1) {
                token_owner = next_ni;  // Assign token to the next PE
                //std::cout <<"The next owner to which the token is given in round-robin  " << next_ni << std::endl;
                token_status = true;   // Mark token as in use
                waiting_set.erase(next_ni);  // Remove from waiting set

                // Remove next_ni from the queue
                waiting_queue.erase(
                    std::remove(waiting_queue.begin(), waiting_queue.end(), next_ni),
                    waiting_queue.end()
                );

                log_info("Token for chiplet " + std::to_string(token_id) + " passed to NI " + std::to_string(next_ni));
            } else {
                // No waiting NIs found
                token_owner = -1;  // No valid PE found
                token_status = false;  // Free the token
                log_info("Token for chiplet " + std::to_string(token_id) + " is now free.");
            }
        } else {
            // No waiting NIs, free the token
            token_owner = -1;
            token_status = false;
            log_info("Token for chiplet " + std::to_string(token_id) + " is now free.");
        }

        return true; // Indicate successful release
    }

    // Utility function to check the current status of the token
    bool get_token_status() const {
        return token_status;
    }

private:
    int token_owner;                // Tracks PE ownership (-1 means no owner)
    bool token_status;              // Token availability
    int last_token_owner;           // To track last owner for round-robin assignment
    std::vector<int> waiting_queue; // Queue for NIs waiting for the token
    std::unordered_set<int> waiting_set; // Set to track NIs already in the queue

    // Print log messages
    void log_info(std::string msg) {
        SC_REPORT_INFO(LOG_NAME, msg.c_str());
    }

    void log_error(std::string msg) {
        SC_REPORT_ERROR(LOG_NAME, msg.c_str());
    }
};

#endif // TOKEN_MANAGER_H
